DEFINT A-Z
'*****************************
'$INCLUDE: 'ndxpack.BI'
'*****************************
'
'  Your application must make room for the recursive procedure calls this
' scheme uses, which after performance reviews appears never to exceed
' 45-50 bytes per call.
'
'Example - StkSpace& = 7500
'          STACK(StkSpace&)                default is 3k
'
'  RECURSIVE procedures in this module are pretty much the only ones that
'  DO NOT carry the STATIC command in their procedure.  In case you
'  haven't figured this one out yet, let me save you some headaches.
'  Putting the STATIC command anywhere in the body of a recursive
'  procedure will cause unpredictable results, to quote that quaint
'  little term so often seen in programming manuals.
'
'  Recursive code is similar to a DO:LOOP structure without a WHILE,UNTIL
'  and instead depends on something inside the loop code to EXIT; and just
'  like this scheme, recursion is very unforgiving.
'
'Example:                   DO
'                            Something1
'                            Something2
'                            PRAY that Something3 happens so you can EXIT
'                           LOOP
'
' Of course, this is a simplistic view (threw that in for you Einsteins)
' since there is the stack and all those other internal bells and whistles to
' contend with, still and all the comparison is good enough for government
' work, so indulge me.
'
' To continue, in order to insure that the something I'm praying for happens,
' all the recursive routines in this scheme are tracked by some internal
' method (see code) and are SUPPOSED to return before the stack vapor-locks.
' The risks of recursive code can't be avoided here; yet, I think I've
' taken sufficient care to minimize the danger.
'
' One way to maximize the stack in a recursive scheme (or any other for that
' matter) is to severely limit the variable assignments within the recursive
' code body.  This I have done through liberal use of COMMON SHARED variables
' whose addresses shouldn't have to be saved on the stack, so study NDXPACK.BI
' carefully for all those variables which pop out of nowhere in ALL my
' procedures, not just the recursive ones.

DEFLNG A-Z
STATIC SUB Build (FileStat%, ErrCode%)
 IF ErrCode% <> 0 THEN EXIT SUB

 'This routine should be called IMMEDIATELY AFTER a call to GETHDR.

 'Check for non-empty index which should not be built

 IF NdxHdr.TRecs <> 0 THEN ErrCode% = 21: EXIT SUB

 'Test for Index & DBFile compatability

 IF UCASE$(NdxHdr.DName) <> UCASE$(DBF.FName) THEN ErrCode% = 19: EXIT SUB

 'Initialize some values

 ErrCode% = 0: Which% = 3: Buf$ = " "

 'Set First Record position, Record length & Record count into variables

 HdrOffset% = DBF.Header.FirstRec
 RecOffset% = DBF.Header.RecLen
 UBnd& = DBF.Header.TRecs
 K& = UBnd&

 'This call is made strictly to get the correct ADDRESS of the INDEX HEADER
 'within the HDX File (NdxAddr&) after checking for valid Index name

 X = GetNdxNum(NdxFile%, NdxHdr.TName, ErrCode%)
 IF ErrCode% <> 0 THEN EXIT SUB

'Okay, here we go.  There will be 3 passes through the DBF file to build
'the index. This first pass will take the first three entries from a range
'which will span the ENTIRE DATABASE (K&=Total records, see RecurRecno).
'this provides the best chance for a nice even balance & spread between
'the root and the first children:
'
'Eg. for values from 1 to 200,           Root = 100
'this is what we'd like to see:         /     \
'                                  Son = 50   Daughter = 200
'
'RecurRecno is the recursive random number generator which will try to find
'the next available record for insertion and returns 0 after Iter% reaches
'some internal value. Remember, the chance that records may be in index
'order in the DBF file necessitates a random record algorithm to prevent
'a linear tendency and recursion/stack problems.
'
'Since it takes 3 passes to index the DBF file, the Insert code (RGIns) has
'been left within this procedure and an integer variable specifies where to
'return after the record picked by RecurRecno has been inserted.

'*******PASS ONE*********************************************************
 FOR I = 1 TO 3
  RANDOMIZE TIMER                           'Seed internal random numb gen.
  Iter% = 0                                 'Initialize iterations
  RecurRecno FileStat%, D1&, Iter%, 0, K&   'Find next available record
  IF D1& <> 0 THEN
   RecNo& = D1&: GOTO RGIns                 'If found one, then do insert
  END IF
Which3:
 NEXT I
'*************************************************************************
'
'Here we are, about to begin PASS 2. This is the workhorse for an Index
'Build.  After PASS1 has finished, most of the records are passed to RGIns
'from here.
'
'Here's where I get a chance to explain why the record buffer is a Common
'Shared String of length 4,000.  To begin with, DBASE III plus allows a
'record buffer to be no more than 4,000 bytes in length.  This I suspect
'comes from DOS' internal buffers for disk access which are by default the
'same size and would maximize the use of read/write buffers (take note: any
'file whose record length evenly divides into 4,000 would be utilizing quite
'an advantage here).
'
'I therefore take advantage of these observations by limiting the range of
'records that PASS2 can insert to always fall within a given subset of
'records that fill the current buffer.  Therefore if the record length is
'80, then the buffer can hold up to 50 records without requiring another
'disk read.
'
'Initialize some variables
 Sector& = 4000 / RecOffset%: LBnd& = 1: Cnt% = 1: Which% = 1

'If there is more room in the buffer than there are records, then step
'through them one at a time.

 IF Sector& > UBnd& THEN Sector& = 1

'*******PASS TWO*************************************************************
											  'Step through the file,
 FOR I& = LBnd& TO UBnd& STEP Sector&         'filling the buffer with recs
  K& = ((Cnt% * Sector&))                     'Set record range to insert
  IF K& > UBnd& THEN K& = UBnd&               'and make sure upper <= UBnd&

  '****Pass 2 inner loop begin***********************************************
  FOR J& = I& TO K&                           'Insert each record in buffer
   RANDOMIZE TIMER                            'Seed internal random numb gen.
   Iter% = 0                                  'Initialize iterations
   RecurRecno FileStat%, D1&, Iter%, I&, K&   'Find next available record
   IF D1& <> 0 THEN
	RecNo& = D1&: GOTO RGIns                  'If found one, then do insert
   END IF
Which1:
  NEXT J&
  '****Pass 2 inner loop end*************************************************
'
' We're still within the DOS BUFFER LOOP (Pass 2). Here is where the third
' pass is done, and as if this isn't confusing enough, it's done in bits and
' pieces; that is, since RecurRecno can and will fail to return all numbers
' within a given range (Speed and accuracy are a trade off, the higher Iter%
' is set to go, the more efficient RecurRecno is, but it will never be 100%
' efficient), we must go through each record in the DOS Buffer manually and
' make sure it is inserted if RecurRecno failed to return that position.
' So . . .
'

  Cnt% = Cnt% + 1                            'Bump count of Dos buffer fills
  Which% = 2                                 'Set RGIns to return to Pass 3

'*******PASS THREE*************************************************************
  FOR L& = I& TO K&                                'Check each rec in buffer
   Test$ = IFlg$(FileStat%, L&, Addr&)             'This code is very
   IF Test$ <> InsFlg$ AND Test$ <> Delete$ THEN   'similar to what Recurecno
	RecNo& = L&: GOTO RGIns                        'does before an insert
   ELSE
	IF Test$ = InsFlg$ THEN               'Here we take advantage to clean up    
	 PUT #FileStat%, Addr&, Buf$          'after Recurecno and restore the
	END IF                                'Delete flag it used for tagging
   END IF                                 'records for insertion with a blank.
Which2:
  NEXT L&
'*******************************PASS THREE***********************************

  Which% = 1                                 'Reset Which to return to PASS 2
 NEXT I&                                     'Resume with next DOS Buffer
'*******************************PASS TWO*************************************

 'Once we're done we can put the Index Header with all of it's information
 'back into the HDX file.

 PUT #NdxFile%, NdxAddr&, NdxHdr

 EXIT SUB
'*************************************************************************
'*************************************************************************
'RGINS
'Subroutine to insert an NdxNode into the Hdx File after RecurRecno or any of
'the Passes has transferred control. There are a couple of internal functions
'here which do the string manipulations and comparisons (GetCmpVal$,CmpValue%)
'and a STKPOP routine which returns the next available spot for the new Node.
'

RGIns:

'Get the record's offset in the DBF file and a comparison string of the
'key values to place into Common Shared Value$

Addr& = HdrOffset% + ((RecNo& - 1) * RecOffset%)
Value$ = UCASE$(GetCmpVal$(FileStat%, Addr&))


IF NdxHdr.Root.RecNo <> 0 THEN
 '******Find a spot in a non-empty index***********************************

 'Get Record's address in DBF file of Root, Assign Node & set a Flag variable

 Addr& = HdrOffset% + ((NdxHdr.Root.RecNo - 1) * RecOffset%)
 Node = NdxHdr.Root
 Flag% = False
 
 DO                                     'Begin search, initially Node = Root

  NewNode.Prnt = Node.Addr                  'Everyone's got a mother, right?

  CmpResult% = CmpValue%(FileStat%, Addr&)  'Which way should I go?

  IF CmpResult% = LeftSide THEN
   Addr& = Node.LChld                                'Hand over to Son
  ELSE
   IF CmpResult% = RightSide AND Same% = False THEN
	 Addr& = Node.RChld                              'Hand over to Daughter
   ELSE
	Addr& = -1                               'Unique=True and Match already
   END IF                                    'in Index.
  END IF

  IF Addr& > 0 THEN                'Still haven't found what we're looking for
   GET #NdxFile%, Addr&, Node      'so get next Node from hand over above,
   Addr& = HdrOffset% + ((Node.RecNo - 1) * RecOffset%)  'and keep looking.
  ELSE
   Flag% = True                    'Bingo!, we can exit the loop, the search
  END IF                           'is over.

 LOOP UNTIL Flag% = True

 IF Addr& = 0 THEN                  'No duplicates were allowed if Addr& = -1
  StkPop Addr&                      'Pop the next available address.
 END IF
ELSE
 '*********Set up the ROOT NODE of the index***************************
 StkPop Addr&                              'Pop the next avail address
 NdxHdr.Root.RecNo = RecNo&                'Root node's Rec. no. from Recur...
 NdxHdr.Root.LChld = 0                     'Null Ptr.
 NdxHdr.Root.RChld = 0                     ' "    "
 NdxHdr.Root.Prnt = 0                      ' "    "
 NdxHdr.Root.Addr = Addr&                  'Save addr returned by StkPop
 NewNode.Prnt = 0                          'Null Ptr.
 PUT #NdxFile%, NdxAddr&, NdxHdr           'Save the Index Header
 '*********************************************************************
END IF

'**********************************************************************
'Finally we can begin the process of Insertion.
'Placing all the information we have gleaned into NewNode for
'writing to the HDX File.
'
IF Addr& > 0 THEN                  'No duplicates were allowed if Addr& = -1

 NewNode.Addr = Addr&              'Setup up newnode.
 NewNode.LChld = 0                 'Remember, it's Prnt was
 NewNode.RChld = 0                 'set above before loop
 NewNode.RecNo = RecNo&            'termination.

 PUT #NdxFile%, Addr&, NewNode     'Write

 IF NewNode.Prnt <> 0 THEN         'Parents always complain. You never call...
								   'Seriously, this is a fix up for the last
								   'Node to take a handoff of NewNode,

  'So, get the parent's Node and Record Address

  GET #NdxFile%, NewNode.Prnt, Node
  Addr& = HdrOffset% + ((Node.RecNo - 1) * RecOffset%)

  'Find out if Parent has a Son or a Daughter

  IF CmpResult% = LeftSide THEN
   Node.LChld = NewNode.Addr
  ELSE
   Node.RChld = NewNode.Addr
  END IF

  PUT #NdxFile%, Node.Addr, Node            'Write Parent back to Hdx file
  IF Node.Addr = NdxHdr.Root.Addr THEN
   NdxHdr.Root.LChld = Node.LChld           'Fix up Index Header if the
   NdxHdr.Root.RChld = Node.RChld           'parent happens to be the Root.
  END IF
 END IF

 NdxHdr.TRecs = NdxHdr.TRecs + 1            'Bump the Index Header Cnt of Recs

END IF

'Once all that is done, RgIns must know where to return.

IF Which% = 1 THEN GOTO Which1 ELSE IF Which% = 3 THEN GOTO Which3 ELSE GOTO Which2

END SUB

DEFINT A-Z
STATIC FUNCTION ChkSum% (Node AS NdxNode)
 'Handy little  function used in Deleting nodes procedures for a
 'quick determination of immediate dependents to a node.

 SELECT CASE Node.LChld
  CASE 0
   IF Node.RChld <> 0 THEN ChkSum% = RightSide ELSE ChkSum% = 0
  CASE ELSE
   IF Node.RChld = 0 THEN ChkSum% = LeftSide ELSE ChkSum% = 3
 END SELECT
END FUNCTION

STATIC FUNCTION CmpValue% (FileStat%, Addr&)
 'All string manipulations and comparisons give programmers headaches sooner
 'or later, and this is it.
 '
 'When handing off Newnodes, Traversing trees, or in general performing any
 'operations on these binary trees, we have to compare string information in
 'order to find out what to do next. DBase DBF files store record information
 'in ASCII TEXT form so I take advantage of this mixed blessing in my compar-
 'ison test to return a Directional value to the caller.
 '
 'Parameters have been pretty straightforward up till now and FileStat% is
 'still the same old tired file handle variable, but;
 '
 'ADDR& is the exact address of the record in the DBF File and must be calc-
 'ulated and passed to this function for use. Common Shared variable Value$
 'must be set to the operand that the Key Value(s) in Record will be compared
 'to.

 'First get the value of Strng$ at Addr& and format it for use.

 Strng$ = UCASE$(GetCmpVal$(FileStat%, Addr&))

 IF LEN(Strng$) > LEN(Value$) THEN
  CmpValue% = RightSide                 'If Strng$ longer then go right
  EXIT FUNCTION
 ELSE
  IF LEN(Strng$) < LEN(Value$) THEN
   CmpValue% = LeftSide                 'If Strng$ shorter then go left
   EXIT FUNCTION
  END IF
 END IF

 IF Strng$ = Value$ THEN                'Else if Strng$ & Value$ match,
  IF NdxHdr.Unq = False THEN            'and Unique flag not set, then
   CmpValue% = RightSide                'by convention, go Right and
   Same = False                         'inform caller of a match by
  ELSE                                  'setting Same to False no less!
   CmpValue% = 0
   Same = True                          'and do opposite if Unique flag set.
  END IF
 ELSE
  '***This is the nontrivial case where most of the work gets done.******
  '                                          Easy part's over, they're not
  Same = False                              'equal but have same length
  IF LTRIM$(Strng$) > LTRIM$(Value$) THEN
   CmpValue% = RightSide                    'so if Strng$ > Value$, go Right
  ELSE
   CmpValue% = LeftSide                     'else go left.
  END IF
 END IF
END FUNCTION

STATIC SUB Convert (Addr&)
 'Pretty straightforward. Take new data elements and format them
 'to prevent wayward address pointers from sneaking through and
 'set the addr of the new element in Hdx file prior to use.
 '
 'This routine is used primarily in deleting nodes procedure and in
 'making a new HDX file.
 '
 NewStkElem.Addr1 = 0
 NewStkElem.Addr2 = 0
 NewStkElem.Addr3 = 0
 NewStkElem.Nxt = 0
 NewStkElem.Addr = Addr&

 NewNode.RecNo = 0
 NewNode.Prnt = 0
 NewNode.LChld = 0
 NewNode.RChld = 0
 NewNode.Addr = Addr&
END SUB

SUB DeleteNdx (NdxHdr AS NdxVol, ErrCode%)
ON LOCAL ERROR GOTO DNdxErr

'First subroutine in project window to directly or indirectly enter a
'RECURSIVE procedure, so pay attention.  If you have already looked at the
'example programs which utilize these procedures, you've seen STKSPACE&
'before with a brief explanation of what it's used for, if not, see the
'header of this module.
'
'Actually, the recursive procedure call, DELTRAV, in this procedure came as
'an afterthought when it became apparent that a standard traverse with the
'operation set to delete was at best unwieldy for the deletion of an entire
'index due to the problem of having to save open file space on the stack.
'
'Standard traversals like the ones in this module work in the following
'manner:
'             1.   Keep going to left/right child till you can't.
'             2.   Perform the operation on the last node you got.
'             3.   Keep going to right/left child till you can't.
'             4.   If you can't go any further, return to parent.
'                 
'Deleting the node means wiping away all of it's information when the delete
'operation is performed in step 2, NO GOOD. So DELTRAV's modus operandi is
'as follows:
'
'             1.   Keep going to left/right child till you can't.
'             2.   Keep going to right/left child till you can't.
'             3.   Perform the delete operation on the last node you got.
'             4.   Return to parent.
'
'Much easier since there's nowhere to go when you delete the node except
'return recursively, now you can save the open file space onto the stack
'after deletion.  See DELTRAV routine for an explanation of how that's done.
'
'SAME is pressed into service here as internal check against Stack overflow.
'
'Formula: MaxCalls = (TotStkSpace / SpaceperRecursion) - Stuffalreadyonstack
'
 Same = (StkSpace& / 48) - 15              '15 actually represents 720 bytes

'Initialize some variables which track Maxcalls(SAME)

 UBnd& = 0: D1& = 0

'Make call to do the deletions.

 Deltrav ErrCode, NdxHdr.Root

'Save Index Header to reflect changes.

 NdxHdr.TRecs = 0
 NdxHdr.Root.RecNo = 0
 PUT #NdxFile, NdxAddr&, NdxHdr

 EXIT SUB

DNdxErr:
 ErrCode% = 15
 EXIT SUB
 RESUME
END SUB

STATIC SUB DelNde (RecNo&, ErrCode%)

'This procedure deletes a single node from the index and will use traverse
'with operation set to delete (32767) for a quick termination after it's
'been found.

'Do some error checking

IF UCASE$(NdxHdr.DName) <> UCASE$(DBF.FName) THEN ErrCode% = 19: EXIT SUB

ErrCode% = 7                  'Huh? All I'm doing is assuming an error.

'Initialize some variables

Node.RecNo = RecNo&: Value$ = ""

'Make the call.

Traverse 1, ErrCode%, 32767

END SUB

SUB Deltrav (ErrCode%, TempNode AS NdxNode)
 DIM TravNode AS NdxNode

'RECURSIVE PROCEDURE!!!!!!!!!!!!!!!!!!!!!!!

 'First do internal checks to avoid exhausting the stack.

 IF ErrCode% <> 0 THEN EXIT SUB
 IF D1& >= Same THEN ErrCode% = 15: EXIT SUB

 'Go to the right till you can't go any further.

 IF TempNode.RChld <> 0 THEN
  D1& = D1& + 1                                'Monitor stack usage
  GET #NdxFile, TempNode.RChld, TravNode
  Deltrav ErrCode%, TravNode                   'Recurse with new node
 END IF

 'Go to the left till you can't go any further.

 IF TempNode.LChld <> 0 THEN
  D1& = D1& + 1                                'Monitor stack usage
  GET #NdxFile, TempNode.LChld, TravNode
  Deltrav ErrCode%, TravNode                   'Recurse with new node
 END IF

 D1& = D1& - 1                  'We're returning to parent, stack will grow.

 Convert (TempNode.Addr)        'Format the last node we got to a NewStkElem.

 StkPut NewStkElem              'Save the node as unused file space.

END SUB

STATIC FUNCTION GetCmpVal$ (FileStat%, Addr&)
'More string manipulations. This internal procedure does the actual Key
'Value(s) extractions from the file. As I said in the header of this module
'Common Shared variables pop out of nowhere in my procedures and this is no
'different; but, timing and sequence is what is crucial in most of the code
'in this scheme, certain things are assumed to be done before you get to a
'certain part, and it is up to the programmer to make sure that they are done
'in order and the proper calls are made before anything is done.
'
'Enough lecturing. I haven't told you anything you don't already know.
'
'This code expects that the Index Header is the correct one, matches the
'DBF file where the info is to be obtained, all error checking has been
'performed, Addr& points to the beginning of the right record, etc.

'Initialize some variables.

 I = 1: Strng$ = ""

 DO

  'Format a Temp. Key variable specified by I within the Index Header

  Buf$ = STRING$(NdxHdr.KeyFld(I - 1).FLen, " ")

  'Calculate the FIELD's position inside the DBF file

  X& = Addr& + NdxHdr.KeyFld(I - 1).FNum - 1

  'Get the field

  GET #FileStat%, X&, Buf$

  'Concatenate a KEY VALUE, removing any leading blanks which will also
  'remove any blank field entries.

  Strng$ = Strng$ + LTRIM$(Buf$)

  'Bump the key field counter

  I = I + 1

 LOOP WHILE I <= NdxHdr.FldCnt

 'Return the result

 GetCmpVal$ = Strng$

END FUNCTION

STATIC SUB GetHdr (FileStat%, NdxNum%, ErrCode%, NdxHdr AS NdxVol)
 ON LOCAL ERROR GOTO GetHdrErr
 ErrCode% = 0

 'NdxHdr and NdxAddr& are Common Shared Variables
 'Set up the Index Header from which all things good come.
 'This pretty much needs to be done whenever an index is going to be
 'accessed/opened for use.  One suggestion I might make here is that
 'you modify the declarations so that NdxHdr is an element in an array
 'of Index Headers which YOU CONTROL(!) allowing you to maintain several
 'indexes at the same time (The recursive procedures herein do not
 'directly access so large a data structure; I'm not that dumb.)

 'Do some error checking.

 IF NdxNum% <= 0 OR NdxNum% > 99 THEN ErrCode% = 16: EXIT SUB

 'Little indirection here, NdxAddr& is used first to store the offset
 'based on NdxNum% in the HDX file header, then to retrieve the actual
 'NdxAddr&

 NdxAddr& = 5 + ((NdxNum% - 1) * 4)
 GET #FileStat%, NdxAddr&, NdxAddr&

 'Get the Index's & the associated DBF's name (See NdxHdr structure)

 GET #FileStat%, NdxAddr& + 2, NdxHdr.TName
 GET #FileStat%, , NdxHdr.DName

 'More error checking.
 IF UCASE$(NdxHdr.DName) <> UCASE$(DBF.FName) THEN ErrCode% = 19: EXIT SUB

 'Proceed to get the rest of the Index Header information

 GET #FileStat%, , NdxHdr.HdrLng
 GET #FileStat%, , NdxHdr.RecLng
 GET #FileStat%, , NdxHdr.Unq
 GET #FileStat%, , NdxHdr.TRecs
 GET #FileStat%, , NdxHdr.Root
 GET #FileStat%, , NdxHdr.FldCnt    'How many index keys are there?

 I = 1                              'Initialize key counter

 DO
  GET #FileStat%, , NdxHdr.KeyFld(I - 1)    'Assign each key header in NdxHdr
  I = I + 1
 LOOP UNTIL I > NdxHdr.FldCnt

 'This assures that subsequent disk reads/writes through Common Shared
 'NdxFile has the correct handle. If you take control of NdxHdr by say,
 'using an array of NdxHdrs, this becomes insufficient.

 NdxFile = FileStat%

 EXIT SUB

GetHdrErr:
 ErrCode% = 9
 EXIT SUB
 RESUME
END SUB

STATIC FUNCTION GetNdxNum% (FileStat%, NdxName$, ErrCode%)
 DIM Index AS STRING * 8
 ON LOCAL ERROR GOTO GetNNErr

 'Search the Hdxfile through the header to find an index matching NdxName$

 'First do some error checking (SIGNATURE)

 GET FileStat%, 969, Num&
 IF Num& <> 969 THEN ErrCode% = 17: EXIT FUNCTION

 'Initialize some variables, assume an error, etc.

 LSET Index = UCASE$(NdxName$)
 GetNdxNum% = 0: Buffer$ = SPACE$(8): ErrCode% = 16

 'Get the count of index headers in the file from the HDX header.

 GET #FileStat%, 3, Cnt

 
 FOR I = 1 TO Cnt                               'Start the search
  GET #FileStat%, 5 + ((I - 1) * 4), Addr&      'Get Index Hdr address
  GET #FileStat%, Addr& + 2, Buffer$            'Get Index name using offset

  IF Index = UCASE$(Buffer$) THEN               'Is it the one?
   GetNdxNum% = I                               'Set return value
   NdxAddr& = 5 + ((I - 1) * 4)           'IMPORTANT! Use indirection first,
   GET #FileStat%, NdxAddr&, NdxAddr&     'then SAVE Index Header ADDR!
   ErrCode% = 0                           'Clear the error setting 
   EXIT FOR
  END IF

 NEXT

 EXIT FUNCTION

GetNNErr:
 ErrCode% = 9
 EXIT FUNCTION
 RESUME
END FUNCTION

STATIC FUNCTION IFlg$ (FileStat%, RecNo&, Addr&)
 'Used to get the Deleted state of a record

 'Initialize some variables

 HdrOffset = DBF.Header.FirstRec - 1: RecOffset = DBF.Header.RecLen
 Buf$ = " "

 'Calculate the record's offset.

 Addr& = HdrOffset + ((RecNo& - 1) * RecOffset) + 1

 'Get record's deleted state and return value.

 GET #FileStat%, Addr&, Buf$
 IFlg$ = Buf$

END FUNCTION

STATIC SUB Key_by_Name (NTemplate$, KeyName$, ErrCode%)

 'This procedure (which calls Key_by_Numb to do the dirty work)
 'assumes that a Template string has been created (see Func NDXTEMPLATE).
 'both Key_by_Numb and ..._Name set bytes in the Keytemplate to create
 'appropriate information for the 10 key fields of a new Index header.

 'Initialze some variables

 ErrCode% = 0: Y = DBF.Header.TFields: Buf$ = UCASE$(KeyName$)

 'Do some error checking : see NDXTEMPLATE for Template string structure.

 IF UCASE$(DBF.FName) <> MID$(NTemplate$, 9, 8) THEN ErrCode% = 19: EXIT SUB

 
 FOR I = 1 TO Y

  'First format the Field name to compare from DBF Field variable.

  Buf1$ = UCASE$(DBF.Fields(I).FName)
  Buf1$ = MID$(Buf1$, 1, INSTR(1, Buf1$, CHR$(0)) - 1)

  'If it's the one we want, save the Field's number in the template & exit.

  IF Buf1$ = Buf$ THEN Key_by_Numb NTemplate$, I, ErrCode%: EXIT FOR

 NEXT I

 'If we didn't find it, set the error code.

 IF I > Y THEN ErrCode% = 25

END SUB

STATIC SUB Key_by_Numb (NTemplate$, KeyNum%, ErrCode%)
ErrCode% = 0

 'This procedure lays down the position of KEY fields in the DBF record
 'as ASCII characters inside the Key template.

 'Do some error checking : see NDXTEMPLATE for Template string structure.

 IF KeyNum% > DBF.Header.TFields THEN ErrCode% = 11: EXIT SUB
 IF LEN(NTemplate$) <> 31 THEN ErrCode% = 23: EXIT SUB
 IF UCASE$(DBF.FName) <> MID$(NTemplate$, 9, 8) THEN ErrCode% = 19: EXIT SUB

 'More error checking : No initial template created or Key area full (>10)
 X = INSTR(21, NTemplate$, CHR$(0))
 IF X = 0 THEN ErrCode% = 23: EXIT SUB
 IF X = 31 THEN ErrCode% = 24: EXIT SUB

 'Set the key field's position in record as an ASCII character.

 MID$(NTemplate$, X, 1) = CHR$(KeyNum%)

END SUB

STATIC SUB MakeHdx (FileStr$, ErrCode%)
  ON LOCAL ERROR GOTO MHdxErr
  ErrCode% = 0

  'This procedure creates default values for an empty HDX file.
  'There may be slight changes here for a NETWORK APP HDX file, so
  'I make not guarantees as to upward mobility. 'Nuff said.

  
  IF FindFile%(FileStr$) = True THEN          'No doppelgangers allowed!
   ErrCode% = 22: EXIT SUB                    '(that's German for evil twin)
  ELSE

   NextFil = FREEFILE: Num& = 969: Mil& = 1000    'Do the work. 
   OPEN FileStr$ FOR BINARY AS NextFil
   Convert (973)
   PUT NextFil, 1, Zero
   PUT NextFil, 3, Zero
   FOR I = 1 TO 99
	PUT NextFil, 5 + ((I - 1) * 4), Zero&
   NEXT
   PUT NextFil, 949, NewStkElem
   PUT NextFil, StkLoc, NewStkElem
   PUT NextFil, 969, Num&
   PUT NextFil, 997, Mil&
   CLOSE NextFil
  END IF

EXIT SUB
MHdxErr:
 ErrCode% = 3: EXIT SUB
 RESUME
END SUB

DEFLNG A-Z
STATIC SUB NdeDel (Node AS NdxNode)
DIM Parent AS NdxNode, Son AS NdxNode, Daughter AS NdxNode
DIM ProbeNode AS NdxNode, PPrnt AS NdxNode, RChld AS NdxNode

'I'm NOT going to be too nice here and explain this algorithm. It took me
'a LONG time to get it to work properly and I'm still a little possessive.
'
'Stop complaining. You've got the code, that's more than I had when I started.
'Let me know if you can think of a better way to do this after you've torn it
'down and built it up again. I really would like to hear from you, and once
'you get a look at the NETWORK APP code for this section, you'll understand
'why.

IF Node.Addr = NdxHdr.Root.Addr THEN
 Flag = True
ELSE
 Flag = False
END IF

IF Flag = False THEN
  GET #NdxFile%, Node.Prnt, Parent
  Children% = ChkSum%(Node)
  SELECT CASE Children%
   CASE 0
	IF Parent.LChld = Node.Addr THEN
	 Parent.LChld = 0
	ELSE
	 Parent.RChld = 0
	END IF
   CASE LeftSide
	GET #NdxFile%, Node.LChld, Son
	IF Parent.LChld = Node.Addr THEN
	 Parent.LChld = Son.Addr
	ELSE
	 Parent.RChld = Son.Addr
	END IF
	Son.Prnt = Parent.Addr
   CASE RightSide
	GET #NdxFile%, Node.RChld, Daughter
	IF Parent.LChld = Node.Addr THEN
	 Parent.LChld = Daughter.Addr
	ELSE
	 Parent.RChld = Daughter.Addr
	END IF
	Daughter.Prnt = Parent.Addr
   CASE ELSE
	GET #NdxFile%, Node.LChld, Son
	GET #NdxFile%, Node.RChld, Daughter
	IF Daughter.LChld = 0 THEN
	 IF Parent.LChld = Node.Addr THEN
	  Parent.LChld = Daughter.Addr
	 ELSE
	  Parent.RChld = Daughter.Addr
	 END IF
	 Daughter.Prnt = Parent.Addr
	 Daughter.LChld = Son.Addr
	 Son.Prnt = Daughter.Addr
	ELSE
	 IF Son.RChld = 0 THEN
	  IF Parent.LChld = Node.Addr THEN
	   Parent.LChld = Son.Addr
	  ELSE
	   Parent.RChld = Son.Addr
	  END IF
	  Son.Prnt = Parent.Addr
	  Son.RChld = Daughter.Addr
	  Daughter.Prnt = Son.Addr
	 ELSE
	  ProbeNode.LChld = Daughter.LChld
	  DO
	   GET #NdxFile%, ProbeNode.LChld, ProbeNode
	  LOOP UNTIL ProbeNode.LChld = 0
	  IF ProbeNode.RChld <> 0 THEN
	   GET #NdxFile%, ProbeNode.RChld, RChld
	  ELSE
	   RChld.Addr = 0
	  END IF
	  GET #NdxFile%, ProbeNode.Prnt, PPrnt
	  IF RChld.Addr = 0 THEN
	   PPrnt.LChld = 0
	  ELSE
	   PPrnt.LChld = RChld.Addr
	   RChld.Prnt = PPrnt.Addr
	  END IF
	  IF Parent.LChld = Node.Addr THEN
	   Parent.LChld = ProbeNode.Addr
	  ELSE
	   Parent.RChld = ProbeNode.Addr
	  END IF
	  ProbeNode.Prnt = Parent.Addr
	  Son.Prnt = ProbeNode.Addr
	  Daughter.Prnt = ProbeNode.Addr
	  ProbeNode.LChld = Son.Addr
	  ProbeNode.RChld = Daughter.Addr
	  PUT #NdxFile%, PPrnt.Addr, PPrnt
	  PUT #NdxFile%, ProbeNode.Addr, ProbeNode
	  IF RChld.Addr <> 0 THEN
	   PUT #NdxFile%, RChld.Addr, RChld
	  END IF
	 END IF
	END IF
  END SELECT

  PUT #NdxFile%, Parent.Addr, Parent
  IF Parent.Addr = NdxHdr.Root.Addr THEN
   NdxHdr.Root = Parent
  END IF
  IF Node.LChld <> 0 THEN
   PUT #NdxFile%, Son.Addr, Son
  END IF
  IF Node.RChld <> 0 THEN
   PUT #NdxFile%, Daughter.Addr, Daughter
  END IF
  NdxHdr.Flag = False
  NdxHdr.TRecs = NdxHdr.TRecs - 1
  PUT #NdxFile%, NdxAddr&, NdxHdr
  Convert Node.Addr
  StkPut NewStkElem
ELSE
 IF NdxHdr.Root.LChld = 0 OR NdxHdr.Root.RChld = 0 THEN
  NdeDelRoot Node
  EXIT SUB
 END IF
 GET #NdxFile%, NdxHdr.Root.LChld, Son
 GET #NdxFile%, NdxHdr.Root.RChld, Daughter
 IF Son.RChld = 0 THEN
  Son.Prnt = 0
  NdxHdr.Root = Son
  NdxHdr.Root.RChld = Daughter.Addr
  Daughter.Prnt = Son.Addr
 ELSEIF Daughter.LChld = 0 THEN
  Daughter.Prnt = 0
  NdxHdr.Root = Daughter
  NdxHdr.Root.LChld = Son.Addr
  Son.Prnt = Daughter.Addr
 ELSE
   ProbeNode.LChld = Daughter.LChld
   DO
	GET #NdxFile%, ProbeNode.LChld, ProbeNode
   LOOP UNTIL ProbeNode.LChld = 0
   IF ProbeNode.RChld <> 0 THEN
	GET #NdxFile%, ProbeNode.RChld, RChld
   ELSE
	RChld.Addr = 0
   END IF
   GET #NdxFile%, ProbeNode.Prnt, PPrnt
   IF RChld.Addr = 0 THEN
	PPrnt.LChld = 0
   ELSE
	PPrnt.LChld = RChld.Addr
	RChld.Prnt = PPrnt.Addr
   END IF
   ProbeNode.LChld = NdxHdr.Root.LChld
   ProbeNode.RChld = NdxHdr.Root.RChld
   ProbeNode.Prnt = 0
   NdxHdr.Root = ProbeNode
   PUT #NdxFile%, PPrnt.Addr, PPrnt
   PUT #NdxFile%, ProbeNode.Addr, ProbeNode
   IF RChld.Addr <> 0 THEN
	PUT #NdxFile%, RChld.Addr, RChld
   END IF
   GOTO Skip_Case_1_2
 END IF
 '******************************
 PUT #NdxFile%, Son.Addr, Son
 PUT #NdxFile%, Daughter.Addr, Daughter
Skip_Case_1_2:
 NdxHdr.Flag = False
 NdxHdr.TRecs = NdxHdr.TRecs - 1
 PUT #NdxFile%, NdxAddr&, NdxHdr
 Convert Node.Addr
 StkPut NewStkElem
END IF
END SUB

DEFINT A-Z
SUB NdeDelRoot (Node AS NdxNode)
DIM Son AS NdxNode, Daughter AS NdxNode

'See disclaimer for NdeDel, from whence and only whence this procedure is
'called.

 IF ChkSum%(NdxHdr.Root) = 0 THEN
  GET #NdxFile, NdxHdr.Root.Addr, Node
  Convert Node.Addr
  StkPut NewStkElem
  NdxHdr.Root.RecNo = 0
  NdxHdr.TRecs = 0
  NdxHdr.Root.Addr = 0
  NdxHdr.Root.LChld = 0
  NdxHdr.Root.RChld = 0
  PUT #NdxFile, NdxAddr&, NdxHdr
  EXIT SUB
 END IF

 IF NdxHdr.Root.RChld = 0 THEN
  GET #NdxFile, NdxHdr.Root.LChld, Son
  Son.Prnt = 0
  NdxHdr.Root = Son
  PUT #NdxFile, Son.Addr, Son
 ELSE
  GET #NdxFile, NdxHdr.Root.RChld, Daughter
  Daughter.Prnt = 0
  NdxHdr.Root = Daughter
  PUT #NdxFile, Daughter.Addr, Daughter
 END IF

 Convert Node.Addr
 StkPut NewStkElem
 NdxHdr.TRecs = NdxHdr.TRecs - 1
 PUT #NdxFile, NdxAddr&, NdxHdr
END SUB

STATIC SUB NdeIns (FileStat%, RecNo&, ErrCode%)
'
'This procedure is virtually identical to RGIns in the BUILD routine
'(sans error checking) so I direct your attention there. Duplication is
'necessary here for portability of code to NETWORK APP code, not just to
'escape error checking. So there.
'
ErrCode% = 0
IF UCASE$(NdxHdr.DName) <> UCASE$(DBF.FName) THEN ErrCode% = 19: EXIT SUB
HdrOffset = DBF.Header.FirstRec
RecOffset = DBF.Header.RecLen
Addr& = HdrOffset + ((RecNo& - 1) * RecOffset)
Value$ = UCASE$(GetCmpVal$(FileStat%, Addr&))
IF NdxHdr.Root.RecNo <> 0 THEN
 Addr& = HdrOffset + ((NdxHdr.Root.RecNo - 1) * RecOffset)
 Node = NdxHdr.Root
 Flag = False
 DO
  NewNode.Prnt = Node.Addr
  CmpResult = CmpValue%(FileStat%, Addr&)
  IF CmpResult = LeftSide THEN
   Addr& = Node.LChld
  ELSE
   IF CmpResult = RightSide AND Same = False THEN
	 Addr& = Node.RChld
   ELSE
	Addr& = -1
   END IF
  END IF
  IF Addr& > 0 THEN
   GET #NdxFile, Addr&, Node
   Addr& = HdrOffset + ((Node.RecNo - 1) * RecOffset)
  ELSE
   Flag = True
  END IF
 LOOP UNTIL Flag = True
 IF Addr& = 0 THEN
  StkPop Addr&
 END IF
ELSE
 StkPop Addr&
 NdxHdr.Root.RecNo = RecNo&
 NdxHdr.Root.LChld = 0
 NdxHdr.Root.RChld = 0
 NdxHdr.Root.Prnt = 0
 NdxHdr.Root.Addr = Addr&
 NewNode.Prnt = 0
 PUT #NdxFile, NdxAddr&, NdxHdr
END IF


IF Addr& > 0 THEN
 NewNode.Addr = Addr&
 NewNode.LChld = 0
 NewNode.RChld = 0
 NewNode.RecNo = RecNo&
 PUT #NdxFile, Addr&, NewNode
 IF NewNode.Prnt <> 0 THEN
  GET #NdxFile, NewNode.Prnt, Node
  Addr& = HdrOffset + ((Node.RecNo - 1) * RecOffset)
  IF CmpValue%(FileStat%, Addr&) = LeftSide THEN
   Node.LChld = NewNode.Addr
  ELSE
   Node.RChld = NewNode.Addr
  END IF
  PUT #NdxFile, Node.Addr, Node
  IF Node.Addr = NdxHdr.Root.Addr THEN
   NdxHdr.Root.LChld = Node.LChld
   NdxHdr.Root.RChld = Node.RChld
  END IF
 END IF
 NdxHdr.TRecs = NdxHdr.TRecs + 1
 PUT #NdxFile, NdxAddr&, NdxHdr
END IF
END SUB

SUB NdeTravA (Operation%, TempNode AS NdxNode)
DIM TravNode AS NdxNode

'RECURSIVE PROCEDURE ASCENDING!!!!!!!!!!!!!!!!!!!!!!!

 'First do internal checks to avoid exhausting the stack,
 'or force a quick return if operation terminated.

 IF Operation% = -1 OR Operation% = -2 THEN EXIT SUB
 IF Same <= D1& THEN Operation = -2: EXIT SUB
 
 IF TempNode.RChld <> 0 THEN                'Go right till you can't anymore
  D1& = D1& + 1                             'monitor stack usage
  GET #NdxFile, TempNode.RChld, TravNode
  NdeTravA Operation%, TravNode             'Recurse with new node.
 END IF

 '**************************************************************************
 'Okay here comes the actual work to be done based on what OPERATION was set
 'to, since this may require stack variables and other messy stuff, I set up
 'Common Shared Variable Strng$ and pass temporary control to NEXTREC.
 'Risky if the stack is bloated, but Nextrec has a STATIC command I
 'recommend you DO NOT fool with and KEEP on any subsequent procedure
 'calls you may make from NEXTREC if called from a recursive procedure.
 'GOT THAT?

 'This is a very long line! (Can you say Supercalifragil....?)

 Strng$ = GetCmpVal$(DBF.Status, DBF.Header.FirstRec + ((TempNode.RecNo - 1) * DBF.Header.RecLen))

 'Go, and keep your fingers crossed (just kidding).

 NextRec DBF.Status, TempNode.RecNo, Strng$, Operation%, TempNode.Addr
 '**************************************************************************

 IF TempNode.LChld <> 0 THEN                'Go left till you can't anymore
  D1& = D1& + 1                             'monitor stack usage
  GET #NdxFile, TempNode.LChld, TravNode
  NdeTravA Operation%, TravNode             'Recurse with new node.
 END IF

 'Planning for disaster is such a pain.  All I'm doing here is keeping track
 'of how many times I tempt fate before Same up above tells me I've gone too
 'far by remembering the LONGEST sequence of consecutive recursions. Let
 'TRAVERSE take care of returning a warning.

 IF UBnd& < D1& THEN UBnd& = D1&

 D1& = D1& - 1                              'Stack is about to grow, return
END SUB                                     'From deepest recursion so far.

SUB NdeTravD (Operation%, TempNode AS NdxNode)
DIM TravNode AS NdxNode

 'RECURSIVE PROCEDURE ASCENDING!!!!!!!!!!!!!!!!!!!!!!!
 'See NdeTravA and DELETENDX for documentation.

 IF Operation% = -1 OR Operation% = -2 THEN EXIT SUB
 IF Same <= D1& THEN Operation = -2: EXIT SUB

 IF TempNode.LChld <> 0 THEN
  D1& = D1& + 1
  GET #NdxFile, TempNode.LChld, TravNode
  NdeTravD Operation%, TravNode
 END IF

 Strng$ = GetCmpVal$(DBF.Status, DBF.Header.FirstRec + ((TempNode.RecNo - 1) * DBF.Header.RecLen))
 NextRec DBF.Status, TempNode.RecNo, Strng$, Operation%, TempNode.Addr
 
 IF TempNode.RChld <> 0 THEN
  D1& = D1& + 1
  GET #NdxFile, TempNode.RChld, TravNode
  NdeTravD Operation%, TravNode
 END IF

 IF UBnd& < D1& THEN UBnd& = D1&
 D1& = D1& - 1
END SUB

STATIC FUNCTION NdxTemplate$ (DBF AS DBFileType, NdxName$, UniqueFlg%)
 DIM Buf AS STRING * 8

 'This sets up a Template String used to create a new index (SETNDX)
 'and for use by Key_by_... routines.
 '
 'At the risk of sounding like David Byrne, you may ask yourself, why did
 'I pass DBF, a Common Shared variable, in the parameter list?  I thought
 'the same thing myself after I'd been away from this piece of code for
 'awhile; but, If you were to set up an array of DBF structures in order
 'to keep track of several DBF's like I suggested for Index Headers in
 'GETHDR, suddenly this starts to make sense. Anyway, you try to figure
 'it out.
 '
 'FORMAT FOR INDEX TEMPLATE STRING:  XXXXXXXXYYYYYYYY----------0
 '
 ' X  - Index Name
 ' Y  - DBF Name
 '   - Ascii representation for DBF Header Length
 '  - Ascii representation for DBF Record Length
 '   - Unique flag
 ' -  > 10 Null fields for Ascii representation of Key field positions
 ' 0  - Terminating Null character
 '
 'Set Unique flag for Template

 IF UniqueFlg% THEN Flg$ = "1" ELSE Flg$ = "0"

 'Set Index name field for Template

 LSET Buf$ = UCASE$(RTRIM$(NdxName$))

 'Set Ascii representations of Header, Record Length

 HdrLen$ = CHR$((DBF.Header.FirstRec - 2) / 32)
 RecLen$ = CHR$(FIX(DBF.Header.RecLen / 254)) + CHR$(DBF.Header.RecLen MOD 254)

 'Concatenate the Template string.

 NdxTemplate$ = Buf$ + UCASE$(DBF.FName) + HdrLen$ + RecLen$ + Flg$ + STRING$(11, 0)

END FUNCTION

STATIC FUNCTION OpenHdx% (FileStr$, ErrCode%)
  ON LOCAL ERROR GOTO OpenHdxErr
  ErrCode% = 0

  'Open an HDX file for use. Mostly just do a quick check to see if file
  'is legit.

  'Do some error checking.

  IF FindFile%(FileStr$) = False THEN
   ErrCode% = 3: EXIT FUNCTION
  ELSE
   NextFil = FREEFILE
  END IF

  'Open the file.

  OPEN FileStr$ FOR BINARY ACCESS READ WRITE SHARED AS NextFil

  'Do some more error checking

  IF LOF(NextFil) < 1000 THEN
   ErrCode% = 17: CLOSE NextFil: EXIT FUNCTION
  ELSE
   GET #NextFil, 969, Test&
   GET #NextFil, 3, Cnt
   IF Test& <> 969 OR Cnt < 0 OR Cnt > 99 THEN ErrCode% = 17: CLOSE NextFil: EXIT FUNCTION
  END IF

  'Return handle of open HDX file.

  OpenHdx% = NextFil

  EXIT FUNCTION

OpenHdxErr:
  IF ERR <> 68 THEN ErrCode% = 5 ELSE ErrCode% = 3
  EXIT FUNCTION
  RESUME
END FUNCTION

DEFLNG A-Z
STATIC FUNCTION Pcnt! ()

 'When I first started coding a random range number generator, I discovered
 'that each seeding of the internal random number generator was unchanging
 'for each application run. The code below seemed to help and consequently
 'improved the divergence of the random pick in a range and hence improved
 'RecurRecno's performance in this area (with smaller values of Iter%), as
 'well as eliminating an annoying preponderance to return the number 1 after
 'rounding and variable conversion, and blah, blah, blah.

 Delta! = RND
 Pcnt! = (Delta! * (10 ^ (INT(5 * Delta! + 1)))) - INT(Delta! * (10 ^ (INT(5 * Delta! + 1))))
END FUNCTION

SUB RecurRecno (FileStat%, D1&, Iter%, I&, K&)

'RECURSIVE PROCEDURE!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

'Given I& and K& as a corresponding LBound and UBound range,
'and Iter% as an internal check for how many times you're going
'to return a number already picked, the rest is just making sure
'you tag the record before you return the recordnumber in D1&
'so you won't waste time returning that number again (See Function IFlg).

'Initialize a variable.

 InsrtFlg$ = "@"

 IF Iter% > K& - I& THEN                'You've tried almost every number
  D1& = 0                               'in the range and gave up. Recursive
 ELSE                                   'returns take care of the rest.

  Iter% = Iter% + 1                            'Another recursion.
  D1& = I& + FIX((K& - I&) * Pcnt!) + 1        'Come on lucky seven!
  RFlg$ = IFlg$(FileStat%, D1&, Addr&)         'Get the flag of Record D1&
  IF RFlg$ = InsFlg$ OR RFlg$ = Delete$ THEN   'Did I win?
   RecurRecno FileStat%, D1&, Iter%, I&, K&    'Craps, pass the dice
  ELSE
   PUT #FileStat%, Addr&, InsrtFlg$            'Winner, tag the record and
  END IF                                       'let recursive returns do
 END IF                                        'the rest.
 'Nota Bene: Addr& is returned by IFlg$ for the file write when a hit is made.
END SUB

DEFINT A-Z
STATIC SUB SetNdx (FileStat%, NTemplate$, NdxNum%, ErrCode%)

 'Okay, so let's assume that all the T's were crossed and the I's dotted
 'before this call was made. Yeah, right. A lot of error checking as we
 'slowly build a new Index header from the Index Template.  If this code
 'is supposed to catch errors, and you decide to maintain several open
 'indexes at once then certain global variables must be restored if an
 'error is caught....
 '
 'By the way, in case I haven't mentioned it yet, this probably is a good
 'way (if not the only way) of deleting an entire index. I don't mean just
 'make the call and everything will work out for crying out loud! First
 'delete the index using DELTRAV which frees up file space and an Index
 'header position. Then make the call AFTER (don't rush me) setting up the
 'New Index Template and placing the New Index Header in the position of
 'the Index you just deleted. If you're thinking to yourself, now he tells
 'me, shame on you for not reading ALL my documentation first. GOTCHA!


 'NdxNum%, the number to assign to the new index, is given a value
 'by the calling sequence.

 'Initialize some files from Common Shared Variables, et al.

 Temp1& = NdxAddr&: Temp2 = NdxFile: ErrCode% = 0

 'Do some error checking: DBF Name, Template length

 IF LEN(NTemplate$) <> 31 THEN ErrCode% = 23: EXIT SUB
 IF UCASE$(DBF.FName) <> MID$(NTemplate$, 9, 8) THEN ErrCode% = 19: EXIT SUB

 'Do some more error checking: Ndx Name and Get Unique Flag
 'also get IndexNumber X% and by so doing NdxAddr&

 Buf$ = MID$(NTemplate$, 1, 8): Flg$ = MID$(NTemplate$, 20, 1)
 X% = GetNdxNum%(FileStat%, Buf$, ErrCode%)
 IF ErrCode% <> 16 THEN GOTO Replace

 SELECT CASE NdxNum%        '*****Let the program assign index number
  CASE 0
   'If for whatever reason we get this far with a duplicate index, EXIT!
   IF X% <> 0 THEN ErrCode% = 26: GOTO Replace

   'Everything looks good so far. Get the count of Indexes in HDX file.
   GET FileStat%, 3, Cnt

   'Check the Count, do what has to be done.
   IF Cnt = 99 THEN ErrCode% = 27: GOTO Replace ELSE Cnt = Cnt + 1

  CASE ELSE                 '*****Write over an existing index.
   'ERROR! Index already exists under number OTHER than suggested by call
   IF X% <> NdxNum% THEN ErrCode% = 16: GOTO Replace

   'Get the number of records for same name & number index
   GET FileStat%, NdxAddr& + 24, TRecs&

   'ERROR! Same name index exists with records in it
   IF TRecs& <> 0 THEN ErrCode% = 21: GOTO Replace

 END SELECT

 'Use the template and DBF to create a header.
 'Mostly pulling teeth and initialization

 X = INSTR(21, NTemplate$, CHR$(0))
 IF X = 21 THEN ErrCode = 28: GOTO Replace
 NdxHdr.Flag = 0
 NdxHdr.TName = Buf$
 NdxHdr.DName = MID$(NTemplate$, 9, 8)
 NdxHdr.HdrLng = (ASC(MID$(NTemplate$, 17, 1)) * 32) + 1
 NdxHdr.RecLng = (ASC(MID$(NTemplate$, 18, 1)) * 254) + ASC(MID$(NTemplate$, 19, 1))
 IF Flg$ = "0" THEN NdxHdr.Unq = False ELSE NdxHdr.Unq = True
 NdxHdr.TRecs = 0
 NdxHdr.Root.Addr = 0: NdxHdr.Root.RecNo = 0
 NdxHdr.FldCnt = 10 - (31 - X)

 'DBF comes into play here

 FOR I = 0 TO NdxHdr.FldCnt - 1
  J = ASC(MID$(NTemplate$, 21 + I, 1))
  NdxHdr.KeyFld(I).FNum = DBF.Fields(J).FOff
  NdxHdr.KeyFld(I).FLen = DBF.Fields(J).FLen
  NdxHdr.KeyFld(I).FType = DBF.Fields(J).FType
 NEXT

 'If this is a NEW index NOT an OVERWRITE, do some fixing up and add the
 'header at the end of the HDX file.

 IF NdxNum% = 0 THEN
  NdxNum% = Cnt
  GET FileStat, FLenAddr, FileLen&
  NdxAddr& = FileLen& + 1
  FileLen& = FileLen& + 100
  PUT FileStat, 3, Cnt
  PUT FileStat, 5 + ((Cnt - 1) * 4), NdxAddr&
  PUT FileStat, 997, FileLen&
 END IF

 'Well what do you know, now we can actually put it to bed,

 PUT FileStat, NdxAddr&, NdxHdr

 'and return the file handle we used to the caller in NdxFile, once again
 'for your benefit if you use an array of NdxHdrs ... you know the rest.
 NdxFile = FileStat%

 EXIT SUB

'What did you forget to do?
'Had to be something otherwise you wouldn't be here.
Replace:
 NdxAddr& = Temp1&: NdxFile = Temp2
END SUB

STATIC SUB StkPop (Addr&)

 'Why'd you go and delete/change information you put here in the first place?
 'Sounds like a personal problem to me.
 'This is where we check the stack to see if some hop-a-long made a booboo
 'and had to delete/edit records from a Dbf file, which affects the Hdx file
 'and so on, before returning an address for the index position of a new or
 'updated (previously deleted, right class?) record.

 'Another quick CAVEAT here, in this single user system, the image of the
 'stack's TAIL is saved in an area RESERVED for the network version of
 'this code, I make no guarantees as to what will happen once you change
 'this code around in any case and this is one of several GOOD reasons why
 'you shouldn't move stuff around too much.

 'Initialize variables, get TAIL of linked list Stack.

 Addr& = 0
 GET #NdxFile, StkLoc, StkElem

 '*****************************************************************
 'Check each of the 3 Stkelem addresses for open space in the HDX file,
 'return in Addr&, Null out the value you took and update the stack.
 'if available space found here.
 IF StkElem.Addr3 <> 0 THEN
  Addr& = StkElem.Addr3
  StkElem.Addr3 = 0
  PUT #NdxFile, StkElem.Addr, StkElem
 ELSEIF StkElem.Addr2 <> 0 THEN
  Addr& = StkElem.Addr2
  StkElem.Addr2 = 0
  PUT #NdxFile, StkElem.Addr, StkElem
 ELSEIF StkElem.Addr1 <> 0 THEN
  Addr& = StkElem.Addr1
  StkElem.Addr1 = 0
  PUT #NdxFile, StkElem.Addr, StkElem
 ELSE
  '*****************************************************************
  'We've got to give up the StkElem itself for use as the new index
  'position address.
  '*****************************************************************

  'Simple enough we've found our volunteer. (Head and tail are not pointing
  'at each other saying "He did it! She did it!)

  IF StkElem.Addr <> StkLoc THEN Addr& = StkElem.Addr

  IF StkElem.Nxt <> 0 THEN
   '****It's not the tail, shorten the linked list.
   GET #NdxFile, StkElem.Nxt, StkElem
   IF Addr& = 0 THEN Addr& = StkElem.Addr1: StkElem.Addr1 = 0
  ELSE
   '****It's the tail, make head and tail reference each other.
   StkElem.Addr = StkLoc
   PUT #NdxFile, 949, StkElem
  END IF

 END IF
 '*****************************************************************

 'You'd lose your head if it wasn't attached to your body.
 PUT #NdxFile, StkLoc, StkElem

 IF Addr& = 0 THEN                          'Did we find an unused address?
  GET #NdxFile, FLenAddr, FileLen&          'No, waste of time, increase LOF
  Addr& = FileLen& + 1                      'Get new addr&
  FileLen& = FileLen& + 20                  'and
  PUT #NdxFile, FLenAddr, FileLen&          'Save the new Filelen
 END IF
END SUB

STATIC SUB StkPut (NewStkElem AS StkType)

 'See STKPOP to watch me blow off some steam for both of us
 'and get a better idea of what's going on here.

 'Get the TAIL

 GET #NdxFile, 949, StkElem

 'Initialize a variable and start looking for the end.
 'But this is the TAIL! Never mind, more confusion with Network Version

 SAddr& = StkElem.Nxt

 IF StkElem.Nxt <> 0 THEN                   'Quick fix and code will still
  DO                                        'make sense to me if not to you.
   GET #NdxFile, SAddr&, StkElem            'so ignore this loop.
   SAddr& = StkElem.Nxt
  LOOP UNTIL SAddr& = 0
 END IF

 'Save the address in an open position & save OR convert it to a new StkElem
 'Either way, replace the tail with updated information.
 IF StkElem.Addr1 = 0 THEN
  StkElem.Addr1 = NewStkElem.Addr
  PUT #NdxFile, 949, StkElem
 ELSEIF StkElem.Addr2 = 0 THEN
  StkElem.Addr2 = NewStkElem.Addr
  PUT #NdxFile, 949, StkElem
 ELSEIF StkElem.Addr3 = 0 THEN
  StkElem.Addr3 = NewStkElem.Addr
  PUT #NdxFile, 949, StkElem
 ELSE
  StkElem.Nxt = NewStkElem.Addr
  PUT #NdxFile, 949, NewStkElem
 END IF

 'Make sure you're able to remember what you did.
 PUT #NdxFile, StkElem.Addr, StkElem
 PUT #NdxFile, NewStkElem.Addr, NewStkElem

END SUB

STATIC SUB Traverse (direction%, ErrCode%, Operation%)
DIM TTempNode AS NdxNode
ON LOCAL ERROR GOTO TravErr

 'Initially all this procedure did was monitor UBnd& (remember him from
 'NDETRAVA & NDETRAVD?) and report a near stack calamity before it got
 'out of hand...

 Same = (StkSpace& / 48) - 15: UBnd& = 0: D1& = 0

 'Later, it also did some error checking...

 IF UCASE$(NdxHdr.DName) <> UCASE$(DBF.FName) THEN ErrCode% = 19: EXIT SUB
 IF NdxHdr.Root.RecNo = 0 THEN ErrCode = 18: EXIT SUB


 'Way back when this was a little baby program, all it did next was call a
 'Recursive Traverse based on direction.
 '
 'Time's change. And it occured to me that we could improve even further
 'on search speed for VERY LARGE DATABASES (VLDB for you lovers of jargon)
 'by specifying a delimiting value based on the traverse direction requested.
 'What's more, the process can be suspended if a match is made (sounds alot
 'like LOCATE in Dbase to me) before a full blown Traverse is made.
 '

 'Initialize some variables.

 TTempNode = NdxHdr.Root: RecOffset = DBF.Header.RecLen: HdrOffset = DBF.Header.FirstRec - 1

 '****************************************************************************
 'I hereby announce my decision to reserve values 1 & 2 of OPERATION. Along
 'with 32767, -1 and -2, that makes 65530 values you can keep (brush up on
 'your algebra if your scratching your head right about now).

 'So if Operation is set to Direction,
 IF Operation% = 1 OR Operation% = 2 THEN

  'And no value was given, exit quietly and leave the programmer puzzled.
  IF Value$ = "" THEN EXIT SUB

  'Else let's move on...
  'What do you know, more formatting and initialization.
  X = LEN(Value$): Value$ = UCASE$(Value$)

  DO

   'Do a quick grab of the Record's key value.
   Address& = HdrOffset + (TTempNode.RecNo - 1) * RecOffset + 1
   Strng$ = UCASE$(MID$(GetCmpVal$(DBF.Status, Address&), 1, X))

   'Quick and dirty exit if the record falls out of a delimiting criteria
   'based on VALUE.
   IF Operation% = 1 AND Strng$ >= Value$ THEN EXIT DO
   IF Operation% = 2 AND Strng$ <= Value$ THEN EXIT DO

   'Otherwise, based on Operation%=Direction%, go find the next position
   'to evaluate.
   IF Operation% = 1 THEN
	IF TTempNode.LChld <> 0 THEN GET #NdxFile, TTempNode.LChld, TTempNode ELSE EXIT SUB
   ELSE
	IF TTempNode.RChld <> 0 THEN GET #NdxFile, TTempNode.RChld, TTempNode ELSE EXIT SUB
   END IF

  LOOP

 END IF
 '****************************************************************************

 'At this point if you spent some time in the area immediately above and
 'got lucky, you could set some flag and fully develop the LOCATE concept.
 'I just move straight into a Traverse.

 SELECT CASE direction%
  CASE 1
   NdeTravA Operation%, TTempNode
  CASE 2
   NdeTravD Operation%, TTempNode
 END SELECT

 'More error checking and fixing on return.
 IF Operation% = -2 THEN ErrCode% = 20: EXIT SUB

 'Remember ErrCode% = 7 initialized in DELNDE?
 IF Operation% = -1 AND ErrCode% = 7 THEN ErrCode% = 0

 'Here's where I check on the State of the Union. If UBnd& comes anywhere
 'near 80% of what Same tells me will max out the stack, I notify the
 'caller: Operation successful but the patient almost died.

 IF UBnd& >= (Same * .8) THEN ErrCode% = 14

 EXIT SUB

TravErr:
 IF ERR = 52 OR ERR = 53 THEN ErrCode% = 3 ELSE ErrCode% = 15
 EXIT SUB
 RESUME
END SUB

